package djmil.cordacheckers.contracts;

import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleInputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.requireThat;

public enum GameProposalCommand implements Command { 
    CREATE,
    ACCEPT,
    REJECT,
    CANCEL;

    public MemberX500Name getInitiator(GameProposalState gameProposalState) {
        switch (this) {
            case CREATE:
            case CANCEL:
                return gameProposalState.getIssuer();
            case ACCEPT:
            case REJECT:
                return gameProposalState.getAcquier();
            default:
                throw new RuntimeException(UNSUPPORTED_VALUE_OF + this.name());
        }
    }

    public MemberX500Name getRespondent(GameProposalState gameProposalState) {
        switch (this) {
            case CREATE:
            case CANCEL:
                return gameProposalState.getAcquier();
            case ACCEPT:
            case REJECT:
                return gameProposalState.getIssuer();
            default:
                throw new RuntimeException(UNSUPPORTED_VALUE_OF + this.name());
        }
    }

    public MemberX500Name getInitiator(StateAndRef<GameProposalState> gameProposalSar) {
        return getInitiator(gameProposalSar.getState().getContractState());
    }

    public MemberX500Name getRespondent(StateAndRef<GameProposalState> gameProposalSar) {
        return getRespondent(gameProposalSar.getState().getContractState());
    }

    public static void validateCreateTrx(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().isEmpty(), CREATE_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, CREATE_OUTPUT_STATE);

        GameProposalState outputState = getSingleOutputState(trx, GameProposalState.class);

        requireThat(outputState.getAcquierColor() != null, NON_NULL_RECIPIENT_COLOR);
    }

    public static void validateAcceptTrx(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, ACCEPT_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, ACCEPT_OUTPUT_STATE);

        GameProposalState inGameProposal = getSingleInputState(trx, GameProposalState.class);
        GameBoardState outGameBoard      = getSingleOutputState(trx, GameBoardState.class);

        requireThat(outGameBoard.getParticipants().containsAll(inGameProposal.getParticipants()), ACCEPT_PARTICIPANTS);
    }

    public static void validateRejectTrx(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, REJECT_INPUT_STATE);
        requireThat(trx.getOutputContractStates().isEmpty(), REJECT_OUTPUT_STATE);

        getSingleInputState(trx, GameProposalState.class);
    }

    public static void validateCancelTrx(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, CANCEL_INPUT_STATE);
        requireThat(trx.getOutputContractStates().isEmpty(), CANCEL_OUTPUT_STATE);
        
        getSingleInputState(trx, GameProposalState.class);
    }

    public static final String UNSUPPORTED_VALUE_OF = "Unsupported GameProposalCommand value: ";

    static final String CREATE_INPUT_STATE  = "Create command should have no input states";
    static final String CREATE_OUTPUT_STATE = "Create command should output exactly one GameProposal state";
    static final String NON_NULL_RECIPIENT_COLOR = "GameProposal.recipientColor field can not be null";

    static final String REJECT_INPUT_STATE  = "Reject command should have exactly one GameProposal input state";
    static final String REJECT_OUTPUT_STATE = "Reject command should have no output states";

    static final String CANCEL_INPUT_STATE  = "Cancel command should have exactly one GameProposal input state";
    static final String CANCEL_OUTPUT_STATE = "Cancel command should have no output states";

    static final String ACCEPT_INPUT_STATE  = "Accept command should have exactly one GameProposal input state";
    static final String ACCEPT_OUTPUT_STATE = "Accept command should have exactly one GameBoard output state";
    static final String ACCEPT_PARTICIPANTS = "Accept command: GameBoard participants should math GameProposal participants";
}
