package djmil.cordacheckers.states;

import java.security.PublicKey;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import djmil.cordacheckers.contracts.GameBoardContract;
import djmil.cordacheckers.states.Piece.Color;
import net.corda.v5.base.annotations.ConstructorForDeserialization;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.BelongsToContract;
import net.corda.v5.ledger.utxo.ContractState;

@BelongsToContract(GameBoardContract.class)
public class GameBoardState implements ContractState, Counterparty {
    
    private final MemberX500Name whitePlayerName;
    private final MemberX500Name blackPlayerName;
    
    private final Piece.Color moveColor;
    private final Map<Integer, Piece> board;
    private final String message;

    private final UUID id;
    private final List<PublicKey> participants;

    public GameBoardState(
        GameProposalState gameBoard
    ) {
        this.whitePlayerName = gameBoard.getWhitePlayerName();
        this.blackPlayerName = gameBoard.getBlackPlayerName();
        
        // Initial GameBoard state
        this.moveColor = Piece.Color.WHITE;
        this.board = new LinkedHashMap<Integer, Piece>(initialBoard);
        this.message = null;

        this.id = UUID.randomUUID();
        this.participants = gameBoard.getParticipants();
    }
 
    @ConstructorForDeserialization
    public GameBoardState(MemberX500Name whitePlayerName, MemberX500Name blackPlayerName, 
            Color moveColor, Map<Integer, Piece> board, String message, 
            UUID id, List<PublicKey> participants) {
        this.whitePlayerName = whitePlayerName;
        this.blackPlayerName = blackPlayerName;
        this.moveColor = moveColor;
        this.board = board;
        this.message = message;
        this.id = id;
        this.participants = participants;
    }

    public MemberX500Name getWhitePlayerName() {
        return whitePlayerName;
    }

    public MemberX500Name getBlackPlayerName() {
        return blackPlayerName;
    }

    public Piece.Color getMoveColor() {
        return moveColor;
    }

    public Map<Integer, Piece> getBoard() {
        return board;
    }

    public String getMessage() {
        return message;
    }

    public UUID getId() {
        return id;
    }

    public List<PublicKey> getParticipants() {
        return participants;
    }

    @Override
    public MemberX500Name getCounterpartyName(MemberX500Name myName) throws NotInvolved {
        if (whitePlayerName.compareTo(myName) == 0)
            return blackPlayerName;
    
        if (blackPlayerName.compareTo(myName) == 0)
            return whitePlayerName;
        
        throw new Counterparty.NotInvolved(myName, GameBoardState.class, this.id);
    }

    public Piece.Color getCounterpartyColor(MemberX500Name myName) throws NotInvolved {
        final MemberX500Name opponentName = getCounterpartyName(myName);

        return getWhitePlayerName().compareTo(opponentName) == 0 ? Piece.Color.WHITE : Piece.Color.BLACK;
    }

    public MemberX500Name getMovePlayerName() {
        switch (moveColor) {
            case WHITE:
                return whitePlayerName;
            case BLACK:
                return blackPlayerName;
            default:
                throw new Piece.Color.UnknownException();
        }
    }

    public final static Map<Integer, Piece> initialBoard =  Map.ofEntries(
        // Inspired by Checkers notation rules: https://www.bobnewell.net/nucleus/checkers.php
        Map.entry( 1, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 2, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 3, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 4, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 5, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 6, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 7, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 8, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry( 9, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry(10, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry(11, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),
        Map.entry(12, new Piece(Piece.Color.BLACK, Piece.Type.MAN)),

        Map.entry(21, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(22, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(23, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(24, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(25, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(26, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(27, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(28, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(29, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(30, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(31, new Piece(Piece.Color.WHITE, Piece.Type.MAN)),
        Map.entry(32, new Piece(Piece.Color.WHITE, Piece.Type.MAN))
        );
}
