package djmil.cordacheckers.gameboard;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleCommand;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;

import java.time.Duration;
import java.time.Instant;
import java.util.Arrays;
import java.util.List;
import java.util.UUID;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.FlowResult;
import djmil.cordacheckers.contracts.GameBoardCommand;
import djmil.cordacheckers.contracts.GameBoardCommand.CommandTypeException;
import djmil.cordacheckers.gameresult.GameResultView;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameResultState;
import djmil.cordacheckers.states.Piece;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.InitiatingFlow;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.application.messaging.FlowMessaging;
import net.corda.v5.application.messaging.FlowSession;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionBuilder;

@InitiatingFlow(protocol = "game-board")
public class CommandFlow implements ClientStartableFlow {

    private final static Logger log = LoggerFactory.getLogger(CommandFlow.class);

    @CordaInject
    public JsonMarshallingService jsonMarshallingService;

    @CordaInject
    public UtxoLedgerService utxoLedgerService;

    @CordaInject
    public MemberLookup memberLookup;

    @CordaInject
    public FlowMessaging flowMessaging;

    @Override
    @Suspendable
    public String call(ClientRequestBody requestBody) {
        log.info("GameBoardCommandFlow started");
        try {
            final CommandFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, CommandFlowArgs.class);

            final StateAndRef<GameBoardState> utxoSarGameBoard = findUtxoGameBoard(args.getGameBoardUuid());
        
            final SecureHash trxId = doTransaction(args.getCommand(), utxoSarGameBoard);
            final Object trxResult = getTrxResult(trxId);

            return new FlowResult(trxResult, trxId)
                .toJsonEncodedString(jsonMarshallingService);
        }
        catch (Exception e) {
            log.warn("GameProposalAction flow failed to process utxo request body " + requestBody + 
                        " because: " + e.getMessage());
            return new FlowResult(e).toJsonEncodedString(jsonMarshallingService);
        }
    }

    @Suspendable
    StateAndRef<GameBoardState> findUtxoGameBoard (UUID gameBoardUuid) {
        /*
         * Get list of all unconsumed aka 'active' GameProposalStates, then filter by UUID.
         * Note, this is an inefficient way to perform this operation if there are a large
         * number of 'active' GameProposals exists in storage.
         */
        return this.utxoLedgerService
            .findUnconsumedStatesByType(GameBoardState.class)
            .stream()
            .filter(sar -> sar.getState().getContractState().getId().equals(gameBoardUuid))
            .reduce((a, b) -> {throw new IllegalStateException("Multiple states: " +a +", " +b);})
            .get();
    }

    @Suspendable
    SecureHash doTransaction(
        GameBoardCommand command,
        StateAndRef<GameBoardState> utxoSarGameBoard
    ) {
        final GameBoardState stateGameBoard = utxoSarGameBoard.getState().getContractState();
        final MemberX500Name myName = memberLookup.myInfo().getName();
        final MemberX500Name opponentName = stateGameBoard.getCounterpartyName(myName);

        UtxoTransactionBuilder trxBuilder = utxoLedgerService.createTransactionBuilder()
            .setNotary(utxoSarGameBoard.getState().getNotaryName())
            .setTimeWindowBetween(Instant.now(), Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
            .addInputState(utxoSarGameBoard.getRef())
            .addCommand(command)
            .addSignatories(stateGameBoard.getParticipants());

        switch (command.getType()) {
            case SURRENDER:
            case ACCEPT:
                final Piece.Color winnerColor = winnerColor(command, stateGameBoard, myName);
                trxBuilder = trxBuilder
                    .addOutputState( new GameResultState(stateGameBoard, winnerColor) );
                break;

            default:
                throw new CommandTypeException("GameBoard.CommandFlow doTransaction()");
        }

        return commit(
            trxBuilder.toSignedTransaction(),
            opponentName);
    }

    @Suspendable
    Piece.Color winnerColor(GameBoardCommand command, GameBoardState stateGameBoard, MemberX500Name myName) {

        switch (command.getType()) {
            case SURRENDER:
                return stateGameBoard.getCounterpartyColor(myName);

            case ACCEPT:
                //stateGameBoard.getMoveColor();
                throw new RuntimeException("Unimplemented");

            default:
                throw new CommandTypeException("GameBoard.CommandFlow winnerColor()");
        }
    }

    @Suspendable
    SecureHash commit(UtxoSignedTransaction candidateTrx, MemberX500Name counterpartyName) {
        log.info("About to commit " +candidateTrx.getId());

        final FlowSession session = flowMessaging.initiateFlow(counterpartyName);

        /*
         * Calls the Corda provided finalise() function which gather signatures from the counterparty,
         * notarises the transaction and persists the transaction to each party's vault.
         */

        final List<FlowSession> sessionsList = Arrays.asList(session);

        final SecureHash trxId = utxoLedgerService
            .finalize(candidateTrx, sessionsList)
            .getTransaction()
            .getId();

        log.info("GameBoard utxo trx id " +trxId);
        return trxId;
    }

    @Suspendable
    Object getTrxResult(SecureHash trxId) {
        final UtxoLedgerTransaction utxoTrx = utxoLedgerService
            .findLedgerTransaction(trxId);

        final var command = getSingleCommand(utxoTrx, GameBoardCommand.class);

        switch (command.getType()) {
            case SURRENDER:
            case ACCEPT:
                return viewGameResult(utxoTrx);

            case MOVE:
            case VICTORY: // request, shall be accepted by opponent
            case DRAW:    // request, shall be accepted by opponent
                return viewGameBoard(utxoTrx);

            default:
                throw new CommandTypeException();
        }
    }

    @Suspendable
    GameResultView viewGameResult(UtxoLedgerTransaction utxoGameResult) {
        final GameResultState grState = getSingleOutputState(utxoGameResult, GameResultState.class);

        return new GameResultView(grState);
    }

    @Suspendable
    GameBoardView viewGameBoard(UtxoLedgerTransaction utxoGameBoard) {
        final MemberX500Name myName = memberLookup.myInfo().getName();

        return new GameBoardView(utxoGameBoard, myName);
    }

}
