package djmil.cordacheckers.gameboard;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleCommand;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleInputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.contracts.GameBoardCommand;
import djmil.cordacheckers.contracts.GameBoardCommand.CommandTypeException;
import djmil.cordacheckers.states.GameBoardState;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.InitiatedBy;
import net.corda.v5.application.flows.ResponderFlow;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.application.messaging.FlowSession;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionValidator;

@InitiatedBy(protocol = "game-board")
public class CommandResponderFlow implements ResponderFlow {
    
    private final static Logger log = LoggerFactory.getLogger(CommandResponderFlow.class);

    @CordaInject
    public MemberLookup memberLookup;

    @CordaInject
    public UtxoLedgerService utxoLedgerService;
    
    @Suspendable
    @Override
    public void call(FlowSession session) {
        try {
            UtxoTransactionValidator txValidator = utxoGameBoard -> {
                final var command = getSingleCommand(utxoGameBoard, GameBoardCommand.class);
                final var stateGameBoard = getGameBoardState(utxoGameBoard, command);

                checkParticipants(session, stateGameBoard, command);

                /*
                 * Other checks / actions ?
                 */

                log.info("Verified the transaction - " + utxoGameBoard.getId());
            };

            UtxoSignedTransaction finalizedSignedTransaction = utxoLedgerService
                    .receiveFinality(session, txValidator)
                    .getTransaction();

            log.info("Finished responder flow - " + finalizedSignedTransaction.getId());
        }
        catch(Exception e) {
            log.warn("Responder flow failed: ", e);
        }
    }

    @Suspendable
    GameBoardState getGameBoardState(UtxoLedgerTransaction utxoGameBoard, GameBoardCommand command) {
        switch (command.getType()) {
            case SURRENDER:
            case ACCEPT:
                return getSingleInputState(utxoGameBoard, GameBoardState.class);

            case MOVE:
            case VICTORY:
            case DRAW:
                return getSingleOutputState(utxoGameBoard, GameBoardState.class); 

            default:
                throw new CommandTypeException();
        }
    }

    @Suspendable
    void checkParticipants(FlowSession session, GameBoardState stateGameBoard, GameBoardCommand command) {        
        final var myName = memberLookup.myInfo().getName();
        final var counterpartyName = session.getCounterparty();

        if (command.getRespondent(stateGameBoard).compareTo(myName) != 0)
            throw new CordaRuntimeException("Bad respondent");

        if (command.getInitiator(stateGameBoard).compareTo(counterpartyName) !=0)
            throw new CordaRuntimeException("Bad initiator");
    }

}
