package djmil.cordacheckers.gameproposal;

import java.util.UUID;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.FlowResult;
import djmil.cordacheckers.contracts.GameProposalCommand;
import djmil.cordacheckers.gameboard.GameBoardView;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.Counterparty.NotInvolved;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionBuilder;

import java.time.Duration;
import java.time.Instant;

public class CommandFlow implements ClientStartableFlow {

    private final static Logger log = LoggerFactory.getLogger(CommandFlow.class);

    @CordaInject
    public JsonMarshallingService jsonMarshallingService;

    @CordaInject
    public UtxoLedgerService utxoLedgerService;

    @CordaInject
    public FlowEngine flowEngine;

    @CordaInject
    public MemberLookup memberLookup;    

    @Override
    @Suspendable
    public String call(ClientRequestBody requestBody) {
        try {
            final CommandFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, CommandFlowArgs.class);
            final GameProposalCommand command = args.getCommand();

            final StateAndRef<GameProposalState> utxoSarGameProposal = findUtxoGameProposal(args.getGameProposalUuid());

            final UtxoSignedTransaction utxoDraft = prepareSignedTransaction(command, utxoSarGameProposal);

            final SecureHash utxoTrxId = this.flowEngine
                .subFlow( new CommitSubFlow(utxoDraft, command.getRespondent(utxoSarGameProposal)) );

            if (command == GameProposalCommand.ACCEPT) {
                final GameBoardView viewGameBoard = prepareGameBoardView(utxoTrxId);

                return new FlowResult(viewGameBoard, utxoTrxId) 
                    .toJsonEncodedString(jsonMarshallingService);
            }

            return new FlowResult(command+"ED", utxoTrxId)  // REJECT+ED
                .toJsonEncodedString(jsonMarshallingService);
        } 
        catch (Exception e) {
            log.warn("GameProposalAction flow failed to process utxo request body " + requestBody + 
                        " because: " + e.getMessage());
            return new FlowResult(e).toJsonEncodedString(jsonMarshallingService);
        }
    }

    @Suspendable
    private StateAndRef<GameProposalState> findUtxoGameProposal (UUID uuidGameProposal) {
        /*
         * Get list of all unconsumed aka 'active' GameProposalStates, then filter by UUID.
         * Note, this is an inefficient way to perform this operation if there are a large
         * number of 'active' GameProposals exists in storage.
         */
        return this.utxoLedgerService
            .findUnconsumedStatesByType(GameProposalState.class)
            .stream()
            .filter(sar -> sar.getState().getContractState().getId().equals(uuidGameProposal))
            .reduce((a, b) -> {throw new IllegalStateException("Multiple states: " +a +", " +b);})
            .get();
    }

    @Suspendable
    private UtxoSignedTransaction prepareSignedTransaction(
        GameProposalCommand command,
        StateAndRef<GameProposalState> sarGameProposal
    ) {
        UtxoTransactionBuilder trxBuilder = utxoLedgerService.createTransactionBuilder()
            .setNotary(sarGameProposal.getState().getNotaryName())
            .setTimeWindowBetween(Instant.now(), Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
            .addInputState(sarGameProposal.getRef())
            .addCommand(command)
            .addSignatories(sarGameProposal.getState().getContractState().getParticipants());

        if (command == GameProposalCommand.ACCEPT) {
            trxBuilder = trxBuilder
                .addOutputState(new GameBoardState(sarGameProposal.getState().getContractState()));
        }

        return trxBuilder.toSignedTransaction();
    }

    @Suspendable
    private GameBoardView prepareGameBoardView(SecureHash utxoTrxId) throws NotInvolved {
        final MemberX500Name myName = memberLookup.myInfo().getName();

        final UtxoLedgerTransaction utxoGameBoard = utxoLedgerService
            .findLedgerTransaction(utxoTrxId);

        return new GameBoardView(utxoGameBoard, myName);
    }

}
