package djmil.cordacheckers.gameproposal;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.contracts.GameProposalCommand;
import djmil.cordacheckers.states.GameProposalState;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.InitiatedBy;
import net.corda.v5.application.flows.ResponderFlow;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.application.messaging.FlowSession;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionValidator;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleCommand;

@InitiatedBy(protocol = "game-proposal")
public class CommitResponderFlow implements ResponderFlow {
    
    private final static Logger log = LoggerFactory.getLogger(CommitResponderFlow.class);

    @CordaInject
    public MemberLookup memberLookup;

    @CordaInject
    public UtxoLedgerService utxoLedgerService;
    
    @Suspendable
    @Override
    public void call(FlowSession session) {
        try {
            UtxoTransactionValidator txValidator = ledgerTransaction -> {
                final GameProposalCommand command = ledgerTransaction.getCommands(GameProposalCommand.class).get(0);

                final GameProposalState gameProposal = getGameProposal(ledgerTransaction); 

                checkParticipants(session, gameProposal, command);

                /*
                 * Other checks / actions ?
                 */

                log.info("Verified the transaction - " + ledgerTransaction.getId());
            };

            UtxoSignedTransaction finalizedSignedTransaction = utxoLedgerService
                    .receiveFinality(session, txValidator)
                    .getTransaction();

            log.info("Finished responder flow - " + finalizedSignedTransaction.getId());
        }
        catch(Exception e) {
            log.warn("Responder flow failed: ", e);
        }
    }

    @Suspendable
    void checkParticipants(
        FlowSession session,
        GameProposalState gameProposal,
        GameProposalCommand command
    ) {
        final MemberX500Name myName = memberLookup.myInfo().getName();
        final MemberX500Name otherName = session.getCounterparty();

        if (command.getRespondent(gameProposal).compareTo(myName) != 0)
            throw new CordaRuntimeException("Bad GameProposal acquirer: expected '"+myName+"', actual '" +gameProposal.getAcquier() +"'");
        
        if (command.getInitiator(gameProposal).compareTo(otherName) != 0)
            throw new CordaRuntimeException("Bad GameProposal issuer: expected '"+otherName+"', actual '"+gameProposal.getIssuer()+"'");
    }

    @Suspendable
    GameProposalState getGameProposal(UtxoLedgerTransaction trx) {
        final GameProposalCommand command = trx.getCommands(GameProposalCommand.class).get(0);
        
        switch (command) {
            case CREATE:
                return (GameProposalState)trx.getOutputContractStates().get(0);

            case ACCEPT:
            case REJECT:
            case CANCEL:
                return (GameProposalState)trx.getInputContractStates().get(0);

            default:
                throw new RuntimeException(GameProposalCommand.UNSUPPORTED_VALUE_OF +command);
        }
    }

}
